/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#ifndef DNS_CACHE_H
#define DNS_CACHE_H

#include <map>

#include <boost/shared_ptr.hpp>
#include <boost/asio/deadline_timer.hpp>
#include <boost/system/error_code.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include "host/pinger.h"    // for IoserviceItem
#include "dns/dnsipprotocol.h"
#include "dns/hostaddress.h"
#include "dns/cname.h"

typedef std::vector<HostAddress> HostAddressVec;
typedef std::pair<std::string, DnsIpProtocol> ip_map_key_type;
typedef std::map<ip_map_key_type, HostAddressVec> ip_map_type;
typedef std::string cname_map_key_type;
typedef std::map<cname_map_key_type, Cname> cname_map_type;

// -----------------------------------------------------------------------------
// DnsCache
// -----------------------------------------------------------------------------

class DnsCache
{
public:
    /// constant to give as cache_file arg to DnsCache constructor
    /// indicating that no cache file should be used
    static const std::string DoNotUseCacheFile;

    DnsCache( const IoServiceItem &io_serv,
              const std::string &cache_file,
              const uint32_t min_time_between_resolves );
    ~DnsCache();

    // accessed from ResolverBase subclasses
    void update(const std::string &hostname,
                const DnsIpProtocol &protocol,
                const HostAddressVec &new_ips);
    void update(const std::string &hostname,
                const Cname &cname);

    // retrieval
    HostAddressVec get_ips(const std::string &hostname,
                           const DnsIpProtocol &protocol,
                           const bool check_up_to_date=false);
    Cname get_cname(const std::string &hostname,
                    const bool check_up_to_date=false);
    HostAddressVec get_ips_recursive(const std::string &hostname,
                                     const DnsIpProtocol &protocol,
                                     const bool check_up_to_date=false);
    std::string get_first_outdated_cname(const std::string &hostname,
                                         const uint32_t ttl_thresh);
    std::string get_cname_chain_str(const std::string &hostname);

// variables
private:
    ip_map_type IpCache;
    cname_map_type CnameCache;
    boost::asio::deadline_timer SaveTimer;
    std::string CacheFile;
    bool HasChanged;
    uint32_t MinTimeBetweenResolves;

// internal functions
private:
    void schedule_save(const boost::system::error_code &error);
    void save_to_cachefile();
    void load_from_cachefile();
    ip_map_key_type key_for_ips(const std::string &hostname,
                                const DnsIpProtocol &protocol) const;
    cname_map_key_type key_for_cname(const std::string &hostname) const;
    bool check_timestamps(const boost::posix_time::ptime &cache_save_time);
    void remove_old_entries();
    void debug_print() const;

};

typedef boost::shared_ptr<DnsCache> DnsCacheItem;

#endif

